/* $Id: platform_win.c 7617 2013-04-26 09:49:01Z chris $ */

#include "platform_win.h"
#include "my_ntddscsi.h"

#define var_offsetof(v, m)                      \
  ((size_t)((char *)&((v).m) - (char *)&(v)))

int open_device(const char *device_name, HANDLE *handle)
{
  HANDLE hVolume;
  UINT uDriveType;
  char szVolumeName[8];
  char szRootName[5];
  DWORD dwAccessFlags;

  if (device_name[0] == '\0' ||
      device_name[1] != ':' ||
      device_name[2] != '\0')
    return EINVAL;

  szRootName[0]=device_name[0];
  szRootName[1]=':';
  szRootName[2]='\\';
  szRootName[3]='\0';

  uDriveType = GetDriveType(szRootName);

  switch(uDriveType) {
  case DRIVE_CDROM:
    dwAccessFlags = GENERIC_READ | GENERIC_WRITE;
    break;

  default:
    return EINVAL;
  }

  szVolumeName[0]='\\';
  szVolumeName[1]='\\';
  szVolumeName[2]='.';
  szVolumeName[3]='\\';
  szVolumeName[4]=device_name[0];
  szVolumeName[5]=':';
  szVolumeName[6]='\0';

  hVolume = CreateFile(   szVolumeName,
                          dwAccessFlags,
                          FILE_SHARE_READ | FILE_SHARE_WRITE,
                          NULL,
                          OPEN_EXISTING,
                          FILE_ATTRIBUTE_NORMAL,
                          NULL );

  if (hVolume == INVALID_HANDLE_VALUE)
    return EINVAL;

  *handle = hVolume;

  return 0;
}

int close_device(HANDLE handle)
{
  BOOL success = CloseHandle(handle);
  return success ? 0 : EINVAL;
}

int read_sector_raw_subchannels(HANDLE handle, size_t sector, size_t sector_count, uint8_t *data)
{
  SCSI_PASS_THROUGH_DIRECT sptd;
  DWORD dwBytesReturned;
  BOOL success;

  sptd.Length=sizeof(sptd);
  sptd.PathId=0;   //SCSI card ID will be filled in automatically
  sptd.TargetId=0; //SCSI target ID will also be filled in
  sptd.Lun=0;      //SCSI lun ID will also be filled in
  sptd.CdbLength=12; //CDB size is 12 for ReadCD MMC1 command
  sptd.SenseInfoLength=0; //Don't return any sense data
  sptd.DataIn=SCSI_IOCTL_DATA_IN; //There will be data from drive
  sptd.DataTransferLength = sector_count * 96;
  sptd.TimeOutValue=60;  //SCSI timeout value (60 seconds - maybe it should be longer)
  sptd.DataBuffer=data;
  sptd.SenseInfoOffset=0;

  //CDB with values for ReadCD CDB12 command.  The values were taken from MMC1 draft paper.
  sptd.Cdb[0]=0xBE;  //Code for ReadCD CDB12 command
  sptd.Cdb[1]=0;

  sptd.Cdb[2] = (sector >> 24) & 0xff;
  sptd.Cdb[3] = (sector >> 16) & 0xff;
  sptd.Cdb[4] = (sector >>  8) & 0xff;
  sptd.Cdb[5] = (sector >>  0) & 0xff;

  sptd.Cdb[6] = (sector_count >> 16) & 0xff;
  sptd.Cdb[7] = (sector_count >>  8) & 0xff;
  sptd.Cdb[8] = (sector_count >>  0) & 0xff;

  sptd.Cdb[9]=0;  // No main channel data
  sptd.Cdb[10]=1;  //Selects read RAW 96 bytes/sector sub-channel data
  sptd.Cdb[11]=0;
  sptd.Cdb[12]=0;
  sptd.Cdb[13]=0;
  sptd.Cdb[14]=0;
  sptd.Cdb[15]=0;

  //Send the command to drive
  success=DeviceIoControl(handle,
                          IOCTL_SCSI_PASS_THROUGH_DIRECT,
                          &sptd, sizeof sptd,
                          &sptd, sizeof sptd,
                          &dwBytesReturned,
                          NULL);

  if (!success ||
      (dwBytesReturned >= (var_offsetof(sptd, ScsiStatus)
                           + sizeof sptd.ScsiStatus) &&
      sptd.ScsiStatus != 0))
    return EIO;

  return 0;
}
